"""Unit test package for bulkdata."""
import os
import sys
 
TEST_DIR = os.path.dirname(os.path.abspath(__file__))
# PROJECT_DIR = os.path.abspath(os.path.join(TEST_DIR, os.pardir))
# sys.path.insert(0, PROJECT_DIR)

BDF_DIR = TEST_DIR + "/bdf"
EXPECT_DIR = TEST_DIR + "/expect"

