"""Top-level package for Bulk Data Python Package."""

__author__ = """Marco T. de Lannoy Kobayashi"""
__email__ = 'mdlkdev@gmail.com'
__version__ = '0.5.2'


from .deck import Deck
from .card import Card